/* gInfo - Copyright (C) 2004 bakunin - Andrea Marchesini 
 *                                      <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "ginfo.h"
#include "ginfo_logo.xpm"

module *module_head = NULL;
module *module_as_default = NULL;
GtkWidget *modules_window = NULL;
GtkWidget *modules_label = NULL;

int modules_c (void);
int modules_xml (void);

int
modules_readdir (char *server)
{
  if (!nobootsplash)
    {
      modules_graphic_init ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      usleep (500000);
    }

  if (server && modules_xml_server (server))
    {
      modules_graphic_close ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      return 1;
    }

  if (modules_c () && !nobootsplash)
    {
      modules_graphic_close ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      return 1;
    }

  if (modules_xml () && !nobootsplash)
    {
      modules_graphic_close ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      return 1;
    }

  if (!nobootsplash)
    {
      modules_graphic_close ();

      while (gtk_events_pending ())
	gtk_main_iteration ();
    }

  return 0;
}

int
modules_c (void)
{
  DIR *d;
  struct dirent *dir;
  struct stat st;
  char *a;

  if (!(d = opendir (GINFO_C_PATH_MODULES)))
    return 1;

  while ((dir = readdir (d)) != NULL)
    {
      int n;

      if (strcmp (dir->d_name + (strlen (dir->d_name) - 3), ".so"))
	continue;

      n = strlen (dir->d_name) + strlen (GINFO_C_PATH_MODULES) + 2;
      if (!(a = (char *) malloc (n * sizeof (char))))
	fatal ("Error: memory.");

      snprintf (a, n, "%s/%s", GINFO_C_PATH_MODULES, dir->d_name);

      if (!lstat (a, &st) && !S_ISDIR (st.st_mode))
	{
	  if (!modules_c_check (a))
	    {
	      if (!nobootsplash)
		modules_graphic_text ("Check module %s", a);
	    }
	  else if (!nobootsplash)
	    modules_graphic_text ("Ignore file %s", a);

	  if (!nobootsplash)
	    {
	      while (gtk_events_pending ())
		gtk_main_iteration ();
	    }
	}

      free (a);
    }

  closedir (d);

  return 0;
}

int
modules_xml (void)
{
  if (modules_xml_dir (GINFO_XML_PATH_MODULES))
    return 1;
  if (modules_xml_dir (modules_xml_tmp_dir (GINFO_TMP_DIR)))
    return 1;

  return 0;
}

int
modules_xml_dir (char *directory)
{
  DIR *d;
  struct dirent *dir;
  struct stat st;
  char *a;

  if (!(d = opendir (directory)))
    return 1;

  while ((dir = readdir (d)) != NULL)
    {
      int n;

      if (strcmp (dir->d_name + (strlen (dir->d_name) - 4), ".xml")
	  && strcmp (dir->d_name + (strlen (dir->d_name) - 5), ".sgml"))
	continue;

      n = strlen (dir->d_name) + strlen (directory) + 2;
      if (!(a = (char *) malloc (n * sizeof (char))))
	fatal ("Error: memory.");

      snprintf (a, n, "%s/%s", directory, dir->d_name);

      if (!lstat (a, &st) && !S_ISDIR (st.st_mode))
	{
	  if (!modules_xml_check (a))
	    {

	      if (!mainwin && !nobootsplash)
		modules_graphic_text ("Check module %s", a);
	    }
	  else if (!mainwin && !nobootsplash)
	    modules_graphic_text ("Ignore file %s", a);

	  if (!nobootsplash)
	    {
	      while (gtk_events_pending ())
		gtk_main_iteration ();
	    }
	}

      free (a);
    }

  closedir (d);

  return 0;
}


void
modules_set_default (char *a, module * m)
{
  if (module_as_default)
    {
      msg ("Error: more one module_as_default in this session: %s\n"
	   "My module_as_default is: %s\n", a, module_as_default->file);
    }
  else
    {
      module_as_default = m;
    }

}

void
modules_graphic_init (void)
{

  GtkWidget *vbox;
  GdkPixmap *pixmap;
  GdkPixmap *mask;
  GtkWidget *image;

  modules_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (modules_window), PACKAGE);
  gtk_window_set_position (GTK_WINDOW (modules_window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (modules_window), FALSE);
  gtk_widget_realize (modules_window);
  gtk_widget_set_size_request (modules_window, 200, 200);


  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (modules_window), vbox);

  pixmap = gdk_pixmap_create_from_xpm_d (modules_window->window, &mask, NULL,
					 ginfo_logo);
  image = gtk_image_new_from_pixmap (pixmap, mask);
  gtk_widget_show (image);

  gtk_box_pack_start (GTK_BOX (vbox), image, TRUE, TRUE, 0);

  modules_label = gtk_label_new ("Test module:");
  gtk_widget_show (modules_label);
  gtk_box_pack_start (GTK_BOX (vbox), modules_label, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (modules_label), GTK_JUSTIFY_CENTER);

  modules_label = gtk_label_new ("initialize data");
  gtk_widget_show (modules_label);
  gtk_box_pack_start (GTK_BOX (vbox), modules_label, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (modules_label), GTK_JUSTIFY_CENTER);

  gtk_widget_show (modules_window);
}

void
modules_graphic_text (char *str, ...)
{

  char buf[100];
  va_list va;
  va_start (va, str);

  vsnprintf (buf, 100, str, va);
  gtk_label_set_text (GTK_LABEL (modules_label), buf);

  va_end (va);
}

void
modules_graphic_close (void)
{
  gtk_widget_destroy (modules_window);
}

void
modules_close (void)
{
  module *t;

  t = module_head;

  while (t)
    {
      module_head = t->next;
      free (t);

      t = module_head;
    }
}

/* Remove functions */
module *
modules_search (char *mod)
{
  module *m = module_head;

  while (m)
    {
      if (!strcmp (m->file, mod))
	return m;

      m = m->next;
    }

  return NULL;
}

void
modules_delete (module * m_del)
{
  module *m, *h = NULL;
  tooltip *t, *k = NULL, *j;
  doc *d;

  m = module_head;

  while (m)
    {
      if (m == m_del)
	{
	  if (!h)
	    module_head = m->next;
	  else
	    h->next = m->next;

	  if (m->type == GINFO_C)
	    {
	      dlclose (m->handle);
	    }
	  else
	    {
	      xmlFree (m->handle);
	      if (m->words)
		free (m->words);
	    }

	  if (module_as_default == m)
	    module_as_default = NULL;

	  t = tooltip_head;

	  while (t)
	    {
	      if (t->module == m)
		{

		  tooltip_number--;

		  if (!k)
		    tooltip_head = t->next;
		  else
		    k->next = t->next;

		  j = t->next;
		  free (t);
		  t = j;
		}
	      else
		{
		  k = t;
		  t = t->next;
		}
	    }

	  while (m->docs)
	    {
	      module_xml_info *i, *f;
	      d = m->docs;

	      i = d->arg;
	      while (i)
		{
		  f = i;
		  i = i->next;
		  free (f);
		}

	      m->docs = m->docs->next;

	      free (d);
	    }

	  free (m);

	  break;
	}

      h = m;
      m = m->next;
    }
}

/* EOF */
