/* gInfo - Copyright (C) 2004 bakunin - Andrea Marchesini 
 *                                      <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "ginfo.h"
#include "ginfo_logo.xpm"

GtkWidget *tutor_win = NULL;
static GtkWidget *tutor_view;
static GtkWidget *tutor_frame;
static GtkWidget *tutor_widget = NULL;
static GdkCursor *ginfo_cursor_link = NULL;
static GdkCursor *ginfo_cursor_normal = NULL;

void tutor_search (GtkWidget *, gpointer);
GtkWidget *tutor_create_doc (void);
void tutor_row_activate_cb (GtkTreeView *, GtkTreePath *,
			    GtkTreeViewColumn *);
void tutor_reset_italic (GtkTreeView * tree, GtkWidget * (*)(void *), void *);
GdkPixbuf *tutor_create_pixbuf (char *);
gboolean ginfo_tutor_xml_visibility_notify_event (GtkWidget *,
						  GdkEventVisibility *,
						  gpointer);
gboolean ginfo_tutor_xml_motion_notify_event (GtkWidget *, GdkEventMotion *,
					      gpointer);
gboolean ginfo_tutor_xml_key_press (GtkWidget *, GdkEventKey *, gpointer);
gboolean ginfo_tutor_xml_event_after (GtkWidget *, GdkEvent *, gpointer);
void ginfo_tutor_xml_check_cursor (GtkTextView *, int, int);
void ginfo_tutor_xml_go_link (GtkTextView *, GtkTextIter *);
doc *ginfo_tutor_xml_search_link (char *);
char *ginfo_tutor_xml_trim (char *);

enum
{
  TITLE_COLUMN,
  FUNC_COLUMN,
  ARG_COLUMN,
  ITALIC_COLUMN,
  NUM_COLUMNS
};

void
tutor_init (void)
{
  GtkWidget *hbox;
  GtkWidget *frame;
  GtkWidget *vbox;
  GtkWidget *sw;
  GtkWidget *vp;
  GtkWidget *entry;
  GtkWidget *sep;
  GtkWidget *label;
  char buf[1024];

  if (tutor_win)
    {
      gtk_widget_destroy (tutor_win);
      return;
    }

  snprintf (buf, 1024, "%s %s - your tutor!", PACKAGE, VERSION);

  tutor_win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_size_request (tutor_win, gdk_screen_width () * 2 / 3,
			       gdk_screen_height () * 2 / 3);
  gtk_window_set_title (GTK_WINDOW (tutor_win), buf);
  gtk_window_set_position (GTK_WINDOW (tutor_win), GTK_WIN_POS_CENTER);

  g_signal_connect ((gpointer) tutor_win, "destroy", G_CALLBACK (tutor_close),
		    NULL);

  frame = gtk_frame_new (buf);
  gtk_widget_show (frame);
  gtk_container_add (GTK_CONTAINER (tutor_win), frame);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_container_add (GTK_CONTAINER (frame), hbox);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, TRUE, 0);

  sep = gtk_hseparator_new ();
  gtk_widget_show (sep);
  gtk_box_pack_start (GTK_BOX (vbox), sep, FALSE, TRUE, 0);

  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (sw);
  gtk_box_pack_start (GTK_BOX (vbox), sw, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  tutor_view = tutor_create_doc ();
  gtk_widget_show (tutor_view);
  gtk_container_add (GTK_CONTAINER (sw), tutor_view);

  sep = gtk_hseparator_new ();
  gtk_widget_show (sep);
  gtk_box_pack_start (GTK_BOX (vbox), sep, FALSE, TRUE, 0);

  label = gtk_label_new ("Search");
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);

  sep = gtk_hseparator_new ();
  gtk_widget_show (sep);
  gtk_box_pack_start (GTK_BOX (vbox), sep, FALSE, TRUE, 0);

  entry = gtk_entry_new ();
  gtk_widget_show (entry);
  gtk_box_pack_start (GTK_BOX (vbox), entry, FALSE, FALSE, 0);

  sep = gtk_hseparator_new ();
  gtk_widget_show (sep);
  gtk_box_pack_start (GTK_BOX (vbox), sep, FALSE, TRUE, 0);

  g_signal_connect ((gpointer) entry, "activate", G_CALLBACK (tutor_search),
		    NULL);

  sep = gtk_vseparator_new ();
  gtk_widget_show (sep);
  gtk_box_pack_start (GTK_BOX (hbox), sep, FALSE, TRUE, 0);

  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (sw);
  gtk_box_pack_start (GTK_BOX (hbox), sw, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  vp = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (vp);
  gtk_container_add (GTK_CONTAINER (sw), vp);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (vp), GTK_SHADOW_NONE);

  tutor_frame = gtk_frame_new (NULL);
  gtk_widget_show (tutor_frame);
  gtk_container_add (GTK_CONTAINER (vp), tutor_frame);

  gtk_widget_show (tutor_win);
}

void
tutor_close (GtkWidget * w, gpointer dummy)
{
  if (tutor_win)
    {
      gtk_widget_destroy (tutor_win);
      tutor_win = NULL;
    }

  if (tutor_widget)
    {
      gtk_widget_destroy (tutor_widget);
      tutor_widget = NULL;
    }
}

void
tutor_search (GtkWidget * w, gpointer dummy)
{
  char *s, *t;

  s = (char *) gtk_entry_get_text (GTK_ENTRY (w));

  if (!s || !*s)
    {
      return;
    }

  t = strdup (s);
  s = t;

  s = trim (s);
  if (!s || !*s)
    {
      free (s);
      return;
    }

  search_char (s);

  free (t);
}

GtkWidget *
tutor_create_doc (void)
{
  GtkTreeSelection *selection;
  GtkCellRenderer *cell;
  GtkWidget *tree;
  GtkTreeViewColumn *column;
  GtkTreeStore *model;
  GtkTreeIter iter;

  module *m = module_head;
  doc *d;

  model =
    gtk_tree_store_new (NUM_COLUMNS, G_TYPE_STRING, G_TYPE_POINTER,
			G_TYPE_POINTER, G_TYPE_BOOLEAN);

  tree = gtk_tree_view_new ();
  gtk_tree_view_set_model (GTK_TREE_VIEW (tree), GTK_TREE_MODEL (model));
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));

  gtk_widget_set_size_request (tree, gdk_screen_width () / 5, -1);

  while (m)
    {
      gtk_tree_store_append (GTK_TREE_STORE (model), &iter, NULL);

      gtk_tree_store_set (GTK_TREE_STORE (model), &iter, TITLE_COLUMN,
			  m->name, FUNC_COLUMN, NULL, ARG_COLUMN, NULL,
			  ITALIC_COLUMN, FALSE, -1);

      d = m->docs;

      while (d)
	{
	  GtkTreeIter child_iter;

	  gtk_tree_store_append (GTK_TREE_STORE (model), &child_iter, &iter);
	  gtk_tree_store_set (GTK_TREE_STORE (model), &child_iter,
			      TITLE_COLUMN, d->text, FUNC_COLUMN, d->func,
			      ARG_COLUMN, d->arg, ITALIC_COLUMN, FALSE, -1);

	  d = d->next;
	}

      m = m->next;
    }

  cell = gtk_cell_renderer_text_new ();

  g_object_set (cell, "style", PANGO_STYLE_ITALIC, NULL);

  column =
    gtk_tree_view_column_new_with_attributes ("Documentation", cell, "text",
					      TITLE_COLUMN, "style_set",
					      ITALIC_COLUMN, NULL);

  gtk_tree_view_append_column (GTK_TREE_VIEW (tree),
			       GTK_TREE_VIEW_COLUMN (column));

  g_signal_connect (tree, "row_activated", G_CALLBACK (tutor_row_activate_cb),
		    model);

  gtk_tree_view_expand_all (GTK_TREE_VIEW (tree));

  return tree;
}

void
tutor_row_activate_cb (GtkTreeView * tree, GtkTreePath * path,
		       GtkTreeViewColumn * column)
{
  GtkTreeIter iter;
  GtkTreeModel *model;
  GtkWidget *(*func) (void *);
  void *arg;

  model = gtk_tree_view_get_model (tree);

  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_model_get (GTK_TREE_MODEL (model), &iter, FUNC_COLUMN, &func,
		      ARG_COLUMN, &arg, -1);

  if (func)
    {
      gtk_tree_store_set (GTK_TREE_STORE (model), &iter, -1);

      tutor_container (func, arg);
    }
}

void
tutor_reset_italic (GtkTreeView * tree, GtkWidget * (*func) (void *),
		    void *arg)
{
  GtkTreeIter iter;
  gboolean italic;
  GtkTreeModel *model;
  GtkTreeIter c_iter;
  GtkWidget *(*f) (void *);
  void *a;
  int k;

  model = gtk_tree_view_get_model (tree);

  gtk_tree_model_get_iter_first (model, &iter);

  do
    {

      k = gtk_tree_model_iter_n_children (model, &iter);

      while (k)
	{
	  gtk_tree_model_iter_nth_child (model, &c_iter, &iter, k - 1);

	  gtk_tree_model_get (GTK_TREE_MODEL (model), &c_iter,
			      ITALIC_COLUMN, &italic, FUNC_COLUMN, &f,
			      ARG_COLUMN, &a, -1);

	  if (italic && (f != func || a != arg))
	    gtk_tree_store_set (GTK_TREE_STORE (model), &c_iter,
				ITALIC_COLUMN, 0, -1);

	  if (f == func && a == arg && !italic)
	    gtk_tree_store_set (GTK_TREE_STORE (model), &c_iter,
				ITALIC_COLUMN, 1, -1);

	  k--;
	}
    }
  while (gtk_tree_model_iter_next (model, &iter));

}

void
tutor_container (GtkWidget * (*func) (void *), void *arg)
{
  GtkWidget *obj;

  if (!func)
    return;

  if (tutor_widget)
    {
      gtk_container_remove (GTK_CONTAINER (tutor_frame), tutor_widget);
    }

  if (!tutor_win)
    tutor_init ();

  obj = func (arg);

  gtk_container_add (GTK_CONTAINER (tutor_frame), obj);
  gtk_widget_show_all (obj);

  tutor_widget = obj;

  tutor_reset_italic (GTK_TREE_VIEW (tutor_view), func, arg);
}

GtkWidget *
tutor_container_default (void *arg)
{
  GtkWidget *w;
  GdkPixmap *pixmap;
  GdkPixmap *mask;
  GtkWidget *image;
  GtkWidget *sep;
  GtkWidget *label;
  char buf[1024];

  if (module_as_default)
    {
      w = module_as_default->docs->func (module_as_default->docs->arg);
      return w;
    }

  w = gtk_vbox_new (FALSE, 0);
  pixmap = gdk_pixmap_create_from_xpm_d (tutor_win->window, &mask, NULL,
					 ginfo_logo);
  image = gtk_image_new_from_pixmap (pixmap, mask);
  gtk_widget_show (image);
  gtk_box_pack_start (GTK_BOX (w), image, TRUE, FALSE, 0);

  sep = gtk_hseparator_new ();
  gtk_widget_show (sep);
  gtk_box_pack_start (GTK_BOX (w), sep, FALSE, TRUE, 0);

  snprintf (buf, 1024, "%s %s\n%s\n%s", PACKAGE, VERSION, AUTHOR, COPYRIGHT);
  label = gtk_label_new (buf);
  gtk_widget_show (label);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_CENTER);
  gtk_box_pack_start (GTK_BOX (w), label, FALSE, TRUE, 0);

  return w;
}

GtkWidget *
ginfo_window (void)
{
  return tutor_win;
}

GtkWidget *
ginfo_tutor_xml (void *arg)
{
  GtkWidget *tv;
  GdkPixbuf *pixbuf;
  GtkTextBuffer *buffer;
  GtkTextTag *tag;
  GtkTextIter iter, iter_start;
  GtkTextMark *start, *end;
  module_xml_info *info = (module_xml_info *) arg;
  module_xml_info_font font;

  int nl = 0;

  memset (&font, 0x0, sizeof (module_xml_info_font));
  font.scale = PANGO_SCALE_MEDIUM;

  tv = gtk_text_view_new ();
  gtk_widget_show (tv);

  buffer = gtk_text_buffer_new (NULL);

  ginfo_cursor_link = gdk_cursor_new (GDK_HAND2);
  ginfo_cursor_normal = gdk_cursor_new (GDK_XTERM);

  gtk_text_view_set_buffer (GTK_TEXT_VIEW (tv), buffer);
  gtk_container_set_border_width (GTK_CONTAINER (tv), 5);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (tv), FALSE);
  gtk_text_view_set_accepts_tab (GTK_TEXT_VIEW (tv), FALSE);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (tv), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (tv), GTK_WRAP_WORD);

  g_signal_connect (G_OBJECT (tv), "key-press-event",
		    G_CALLBACK (ginfo_tutor_xml_key_press), tv);
  g_signal_connect (G_OBJECT (tv), "event-after",
		    G_CALLBACK (ginfo_tutor_xml_event_after), tv);
  g_signal_connect (G_OBJECT (tv), "motion-notify-event",
		    G_CALLBACK (ginfo_tutor_xml_motion_notify_event), tv);
  g_signal_connect (G_OBJECT (tv), "visibility-notify-event",
		    G_CALLBACK (ginfo_tutor_xml_visibility_notify_event), tv);

  gtk_text_buffer_get_iter_at_offset (buffer, &iter, 0);
  start = gtk_text_buffer_create_mark (buffer, "start", &iter, 1);
  end = gtk_text_buffer_create_mark (buffer, "end", &iter, 1);

  while (info)
    {
      gtk_text_buffer_move_mark (buffer, start, &iter);

      switch (info->type)
	{

	case INFO_TEXT:
	  {
	    char *last_trim;

	    if (nl)
	      last_trim = ginfo_tutor_xml_trim (info->text);
	    else
	      last_trim = info->text;

	    nl = 0;

	    gtk_text_buffer_insert (buffer, &iter, last_trim,
				    strlen (last_trim));
	    break;
	  }

	case INFO_NEWLINE:

	  nl = 1;

	  gtk_text_buffer_insert (buffer, &iter, "\n", 1);
	  break;

	case INFO_SPACE:
	  gtk_text_buffer_insert (buffer, &iter, " ", 1);
	  break;

	case INFO_IMAGE:
	  if ((pixbuf = tutor_create_pixbuf (info->image)))
	    gtk_text_buffer_insert_pixbuf (buffer, &iter, pixbuf);
	  break;

	case INFO_FONT:
	  memcpy (&font, &info->font, sizeof (module_xml_info_font));
	  break;
	}

      gtk_text_buffer_move_mark (buffer, end, &iter);

      gtk_text_buffer_get_iter_at_mark (buffer, &iter_start, start);

      /* LINK */
      if (info->link)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "foreground",
					"blue", "underline",
					PANGO_UNDERLINE_SINGLE, NULL);

	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);

	  g_object_set_data (G_OBJECT (tag), "link", info->link);
	}

      /* FONT */
      if (font.style != PANGO_STYLE_NORMAL)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "style", font.style,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.variant != PANGO_VARIANT_NORMAL)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "variant",
					font.variant, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.weight != PANGO_WEIGHT_NORMAL)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "weight",
					font.weight, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.stretch != PANGO_STRETCH_NORMAL)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "stretch",
					font.stretch, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.scale != PANGO_SCALE_MEDIUM)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "scale", font.scale,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* SIZE */
      if (font.size)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "size", font.size,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* COLOR */
      if (font.foreground)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "foreground",
					font.foreground, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.background)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "background",
					font.background, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* JUSTIFY */
      if (font.justify)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "justification",
					font.justify, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* UNDERLINE */
      if (font.underline == 1)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "underline",
					PANGO_UNDERLINE_SINGLE, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      if (font.underline == 2)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "underline",
					PANGO_UNDERLINE_DOUBLE, NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      /* FACE */
      if (font.face)
	{
	  tag =
	    gtk_text_buffer_create_tag (buffer, NULL, "font", font.face,
					NULL);
	  if (tag)
	    gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);
	}

      tag =
	gtk_text_buffer_create_tag (buffer, NULL, "left_margin",
				    font.margin_left, "right_margin",
				    font.margin_right, NULL);
      if (tag)
	gtk_text_buffer_apply_tag (buffer, tag, &iter_start, &iter);

      info = info->next;
    }

  return tv;

}

GdkPixbuf *
tutor_create_pixbuf (char *image)
{
  struct stat st;
  GdkPixbuf *pixbuf;
  int n;
  char *f;

  if (!image)
    return NULL;

  n = strlen (GINFO_XML_PATH_MODULES) + strlen (image) + 2;
  if (!(f = (char *) malloc (sizeof (char) * n)))
    fatal ("Error: memory.");
  snprintf (f, n, "%s/%s", GINFO_XML_PATH_MODULES, image);

  if (!lstat (f, &st))
    {
      if ((pixbuf = gdk_pixbuf_new_from_file (f, NULL)))
	{
	  free (f);
	  return pixbuf;
	}

      fprintf (stderr, "I cant' convert your %s file.\n", f);
      free (f);
      return NULL;
    }

  free (f);

  if (!lstat (image, &st))
    {
      if ((pixbuf = gdk_pixbuf_new_from_file (image, NULL)))
	return pixbuf;

      fprintf (stderr, "I cant' convert your %s file.\n", image);
      return NULL;
    }

  fprintf (stderr, "Image %s not found\n", image);
  return NULL;
}

gboolean
ginfo_tutor_xml_motion_notify_event (GtkWidget * w,
				     GdkEventMotion * event, gpointer dummy)
{
  gint x, y;
  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (dummy),
					 GTK_TEXT_WINDOW_WIDGET, event->x,
					 event->y, &x, &y);

  ginfo_tutor_xml_check_cursor (GTK_TEXT_VIEW (dummy), x, y);
  gdk_window_get_pointer (GTK_WIDGET (dummy)->window, NULL, NULL, NULL);

  return FALSE;
}

gboolean
ginfo_tutor_xml_visibility_notify_event (GtkWidget * w,
					 GdkEventVisibility * event,
					 gpointer dummy)
{
  gint wx, wy, bx, by;

  gdk_window_get_pointer (GTK_WIDGET (dummy)->window, &wx, &wy, NULL);

  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (dummy),
					 GTK_TEXT_WINDOW_WIDGET, wx, wy,
					 &bx, &by);

  ginfo_tutor_xml_check_cursor (GTK_TEXT_VIEW (dummy), bx, by);
  gdk_window_get_pointer (GTK_WIDGET (dummy)->window, NULL, NULL, NULL);

  return FALSE;
}

gboolean
ginfo_tutor_xml_key_press (GtkWidget * w, GdkEventKey * event, gpointer dummy)
{
  GtkTextIter iter;
  GtkTextBuffer *buffer;

  switch (event->keyval)
    {
    case GDK_Return:
    case GDK_KP_Enter:
      buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (dummy));
      gtk_text_buffer_get_iter_at_mark (buffer, &iter,
					gtk_text_buffer_get_insert (buffer));
      ginfo_tutor_xml_go_link (GTK_TEXT_VIEW (dummy), &iter);
      break;

    default:
      break;
    }

  return FALSE;

}

gboolean
ginfo_tutor_xml_event_after (GtkWidget * w, GdkEvent * event, gpointer dummy)
{
  GtkTextIter start, end, iter;
  GtkTextBuffer *buffer;
  GdkEventButton *e;
  gint x, y;

  if (event->type != GDK_BUTTON_RELEASE)
    return FALSE;

  e = (GdkEventButton *) event;

  if (e->button != 1)
    return FALSE;

  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (dummy));

  gtk_text_buffer_get_selection_bounds (buffer, &start, &end);
  if (gtk_text_iter_get_offset (&start) != gtk_text_iter_get_offset (&end))
    return FALSE;

  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (dummy),
					 GTK_TEXT_WINDOW_WIDGET, e->x, e->y,
					 &x, &y);

  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (dummy), &iter, x, y);
  ginfo_tutor_xml_go_link (GTK_TEXT_VIEW (dummy), &iter);

  return FALSE;
}

void
ginfo_tutor_xml_check_cursor (GtkTextView * tv, int x, int y)
{
  GSList *tags = NULL, *tagb = NULL;
  GtkTextBuffer *buffer;
  GtkTextIter iter;
  int do_link = 0;

  buffer = gtk_text_view_get_buffer (tv);
  gtk_text_view_get_iter_at_location (tv, &iter, x, y);
  tags = tagb = gtk_text_iter_get_tags (&iter);

  while (tags)
    {
      GtkTextTag *tag = tags->data;
      char *link = (char *) g_object_get_data (G_OBJECT (tag), "link");

      if (link)
	do_link = 1;

      tags = tags->next;
    }

  if (do_link)
    gdk_window_set_cursor (gtk_text_view_get_window
			   (tv, GTK_TEXT_WINDOW_TEXT), ginfo_cursor_link);
  else
    gdk_window_set_cursor (gtk_text_view_get_window
			   (tv, GTK_TEXT_WINDOW_TEXT), ginfo_cursor_normal);

  if (tagb)
    g_slist_free (tagb);
}

void
ginfo_tutor_xml_go_link (GtkTextView * tv, GtkTextIter * iter)
{
  GSList *tags = NULL, *tagb = NULL;

  tags = tagb = gtk_text_iter_get_tags (iter);

  while (tags)
    {
      GtkTextTag *tag = tags->data;
      char *link = g_object_get_data (G_OBJECT (tag), "link");

      if (link != 0)
	{
	  doc *d = ginfo_tutor_xml_search_link (link);

	  if (d)
	    tutor_container (d->func, d->arg);
	  else
	    msg ("I can't find %s link!", link);

	  break;
	}

      tags = tags->next;
    }

  if (tagb)
    g_slist_free (tagb);
}

doc *
ginfo_tutor_xml_search_link (char *link)
{
  module *m = module_head;
  doc *d;

  while (m)
    {
      d = m->docs;

      while (d)
	{
	  if (d->id)
	    if (!strcmp (link, d->id))
	      return d;

	  d = d->next;
	}
      m = m->next;
    }

  return NULL;
}

char *
ginfo_tutor_xml_trim (char *text)
{
  int i = 0, len;
  int j = 0;

  len = strlen (text);

  while (i < len)
    {
      if (text[i++] != ' ')
	break;
      else
	j++;
    }

  return &text[j];
}

/* EOF */
